%% Supplamentay material to the paper:
% Linear Characteristic Graphical Models: Representation, Inference and
% Applications. By Danny Bickson and Carlos Guestrin. In NIPS 2010.
% http://arxiv.org/abs/1008.5325
% 
% 
% Written by Danny Bickson, CMU
% Source code is available from http://www.cs.cmu.edu/~bickson/stable/
%
%
% function stable_jacobi - computes iterative inference a linear stable
% model. See algorithm 2(c) in the above paper.
% inputs: 
% J - invertible matrix
% s - observations (cell array with stable distribution characterized by
% the four parameters alpha beta gamma and delta in each cell position)
% iter - number of iterations to run
% tol - tolerance
% prior - stable prior.
%
% Output:
% x - cell array of size length(s) with stable distributions in each
% position
% bnorm, gnorm, dnorm - relative norms of beta, gamma and delta
% G,BB,D - history of values of gamma, beta and delta
% r - number of iterations before convergence


function [x,bnorm,gnorm,dnorm,G,BB,D,r] = stable_jacobi2a(J, s, iter, tol, prior)
    
    n = length(J);

    x = cell(n,1);
    elem= prior;
    betay = zeros(n,1);
    gammay = zeros(n,1);

    G=zeros(iter,n); D=G; BB=G;
    bnorm=zeros(iter,1); gnorm = bnorm; dnorm = bnorm;
    
    for i=1:n
        x{i} = elem;
        betay(i) = s{i}(2);
        gammay(i) = s{i}(3);
    end
    assert(min(gammay) > 0);
    assert(max(betay) <= 1);
    assert(min(betay) >= -1);
 
    alpha = prior(1);
    assert(alpha>0 && alpha<=2);
    
    gamma = zeros(n,1);
    beta = zeros(n,1);
    delta = zeros(n,1);
    
    xi = zeros(n,1);
    
    for r=1:iter
    
        disp(['iter ', num2str(r)]);
        for i=1:n
             gamma(i) = s{i}(3)^alpha - prior(3)^alpha;
             beta(i) = s{i}(2) * s{i}(3)^alpha - prior(2) * prior(3)^alpha;
             delta(i) = s{i}(4) - prior(4) - xi(i);
             %assert(isreal(delta(i)));
             
             for j=1:n
                 
                 if (i~=j && J(i,j)~=0)
                      gamma(i) = gamma(i) - abs(J(i,j))^alpha*gamma(j);
                      beta(i) = beta(i) - abs(J(i,j))^alpha*sign(J(i,j))*beta(j);
                      delta(i) = delta(i) - J(i,j)*delta(j);
                 end
                 
                 
             end
             
             G(r,i) = gamma(i); BB(r,i) = beta(i); D(r,i) = delta(i);
            
              
        end
        
        if (r > 1)
           bnorm(r) = norm(old_beta - beta);
           gnorm(r) = norm(old_gamma -gamma);
           dnorm(r) = norm(old_delta - delta);
           disp(['b=norm ', num2str(bnorm(r)), ' g=norm ', num2str(gnorm(r)), ' d=norm ', num2str(dnorm(r))]);
           if (bnorm(r) < tol && gnorm(r) < tol && dnorm(r) < tol)
                disp(['Jacobi converged in round ', num2str(r)]);break;
           end
        end
        
        
        old_beta = beta; old_gamma = gamma; old_delta = delta;
        
         tbeta = beta./gamma;
         tgamma = gamma.^(1/alpha);
             
         if (alpha ~= 1)
                xi =  tan(0.5*pi*alpha)*(betay.*gammay - J*(tbeta.*tgamma)); 
         else
            B=J.*log(abs(J));
            B(J==0)=0; 
           
            assert(min(tgamma) > 0);
            xi = (2/pi)*(betay.*gammay.*log(gammay) - B*(tbeta.*tgamma) - J*(tbeta.*tgamma.*log(tgamma)));
         end
        
    end
    
    beta = beta./gamma;
    gamma = gamma.^(1/alpha);
    

    for i=1:n
       x{i}(3) = gamma(i);
       x{i}(2) = beta(i);
       x{i}(4) = delta(i);
    end
    
end
